/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     coded in ECMAScript 5 - JavaScript 5  (ES5, ECMAScript 2009)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


//  xxxxxxxxxxxxxxx   wm13 variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 

var itemCnt = 3;       // total item count 

var exact = false;     // exact solution must exist (true) or there must be only an approximate solution (false)

var autocompute = true;   // Automatic calculation is on (true) or off (false)

var is_int = false;     // true => The final result of an arithmetic expression must be integer. false => The final result of an arithmetic expression can be a decimal number or an integer

var is_req = false;   // true => The user must enter a valid arithmetic expression before moving on to the next task. false => he user does not have to enter a valid answer to the task 

var warning = false;  // Go to the next task - must be confirmed or not

var demo = true;      // if (demo == true) => correct solution(s) is shown

var solCnt = 10;      // number of optimal solutions shown (arithmetic expressions)

var timer_enabled = true;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 240;     // Task time in seconds if clock = 1000ms (default). 

var prg_type = "1";   //  The type of information displayed on the Progress bar ( "0" == Items/Total "1"= Time, "2"= Nothing )

var store = false;    // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other wm13 variables and constants:
   
var clock = 1000;  // basic timer resolution in ms. Default = 1s (1000ms)

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd()   

var item = 0;   // current item      

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;    // task in progress or not  (for timer)
      
var result = {};  // final filtered results (only different numeric values) 

var digits = [1, 2, 3, 4, 5];   // digits array
 
/*  dv = [] = dependent variable for each item=> two-dimensional array (matrix). 
    2. index explanation:  
    [0] => true => arithmetic expression is valid;  false => arithmetic expression is not ok
		[1] => user entered arithmetic expression (string)
		[2] => user result of arithmetic expression
		[3] => The best (optimal) result
		[4] => Accuracy - percentage of similarity (absolute value) with the best solution
		[5] => Error percentage (deviation - absolute value from the optimal result)	*/
var dv = [];	

var target_num;  // target number (it changes for each task)

var optimal = '';   // best (optimal) arithmetic expression for the current item (string)


//  xxxxxxxxxxxxxxx     wm13 functions  xxxxxxxxxxxxxxxx


// init html page [create/start Timer, check storage, show start panel..] 
function init() 
{ 
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {wm13Timer()},clock);      
  
	if (!supportedTransform())  {
    doCenter(document.getElementById("start"));
  } 
	
	if (isStorage() && storageAvailable('localStorage'))  {
    document.getElementById("store_group").disabled= false;
  }
  else  {
    document.getElementById("store_msg").textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline).";
    document.getElementById("store_group").disabled= true;  
    document.getElementById("clr_store").disabled= true;     // IE11 bug  => childrens are not auto disabled
    document.getElementById("store").disabled= true;          // IE11 bug => childrens are not auto disabled
  }          
	document.getElementById('answer').onkeypress = answerKeyPress;	
	document.getElementById('answer').onkeyup = answerKeyUp;	
	document.getElementById('answer').onpaste= function(){return false}; // Prevent pasting in answer input type=text (since pasted content might include non-allowed characters)
	//document.getElementById('answer').onpaste = event => false; // work in ECMAScript 2015
	setIns();
	showStart(); 
}


//  check for browser supported transform
function supportedTransform() 
{
  var prefixes = "transform WebkitTransform MozTransform OTransform msTransform".split(" ");
  var div = document.getElementById("wrapper");
  for(var i = 0; i < prefixes.length; i++) {
    if(div && div.style[prefixes[i]] !== undefined) {
      return prefixes[i];
    }
  }
  return false;
}


// center child panel(s)
function doCenter(e)  
{      
  e.style= (window.innerWidth > 600 && window.innerHeight > 600 ) 
  ? "position:absolute;min-width:50%;min-height:50%;margin:auto;overflow:auto;"  
  : "position:absolute;min-width:75%;min-height:75%;margin:auto;overflow:auto;";  
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== "undefined") {
     return true;               //  localStorage/sessionStorage is supported     
  } 
  else {
    return false;           // no browser Web Storage support
  }  
}


// necessarily because of IE11 and Edge offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// set instructions text based on application options
function setIns()  
{
  document.getElementById("instructions").style.display= "block";			
	document.getElementById("ins_int").style.display= is_int ? "block" : "none";	
	if (document.getElementById("timer_enabled").checked)  
	{
		document.getElementById("ins_timer").setAttribute("style", "display:block");
		document.getElementById("ins_timer").textContent= "The duration of this task is limited.";  // The duration of this task is always limited.
		if (prg_type === "0")  {
	    document.getElementById("ins_timer").textContent+=" The Progress bar will show how many items are left until the end of the task.";
		}
		else if (prg_type === "1")  {
		  document.getElementById("ins_timer").textContent+= " The Progress bar will display the remaining time until the end of the task.";  
		}
		else  {
		  document.getElementById("ins_timer").textContent+= " But information about the remaining time will not be visible";  
		}  
  }
	else  {
		document.getElementById("ins_timer").setAttribute("style", "display:none");
	}			
	document.getElementById("ins_calc").style.display= autocompute ? "block" : "none";		
	document.getElementById("ins_demo").style.display= demo ? "block" : "none";	 
}	


// show start panel 
function showStart()  
{      
  document.getElementById("end").style.display= "none";
  document.getElementById("result_wrapper").style.display="none";
  document.getElementById("table_grp").style.display="none";  
  document.getElementById("start").style.display= "block";
}


// show forms with application options (settings)
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById("hide_option").style.display= "block";
  document.getElementById("options").style.display= "block";
  document.getElementById("start").classList.remove("child");  // don't work in <= IE 9    
  document.getElementById("start_div").style.display="none";  // .disabled = true;   
  document.getElementById("show_option").style.display ="none";  
  document.getElementById("hide_option").setAttribute("style","text-align:center");  // bug? auto-remove style when element.style.display="none"
  document.getElementById("instructions").style.display ="none";     
  if (document.getElementById("store_group").disabled === false && localStorage.getItem('wm13-v1.0')) 
  {
    var cnt= JSON.parse(localStorage.getItem('wm13-v1.0')).length;                                                        
    document.getElementById("store_cnt").textContent=  'You have a total of ' + cnt  + ' wm13 results stored in your web browser';
  }
  dirty= false;
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  
	document.getElementById("options").style.display="none"; 
  document.getElementById("start").classList.add("child");   
  document.getElementById("show_option").style.display ="block";   
  document.getElementById("start_div").style.display="block";        //.removeAttribute("disabled");  
  document.getElementById("hide_option").style.display ="none";   
  document.getElementById("start_div").setAttribute("style","text-align:center");  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById("show_option").setAttribute("style","text-align:center");   // bug?

  setIns();  
      
  refMsg();  
}

       
// refresh variable research Options (update *_msg label) => give user short option HELP
function refOption(e)  
{   
  dirty=true;       
  var my_id = e.id, msg = document.getElementById(my_id + "_msg"), s = 'Values within this range are allowed: ', s2= ' with step= ';    
  
  switch (my_id)
  {
    
    case "item_cnt":
		case "sol_cnt":
			if (e.validity.valid)  {           
				msg.removeAttribute("class");       
				msg.textContent= '';       //msg.textContent=  e.value; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute("class", "error");    // msg.className = 'error';
			}       		
			break;						                      
  					
    case "exact":                                                                                                                                       
      if (e.value === "1")  {
        msg.textContent= 'It is always possible to generate an arithmetic expression whose result is exactly equal to the target number.';
      }
      else  {
        msg.textContent= 'It is possible that the target number cannot always be calculated using 5 digits. In these cases, only approximate task solutions will exist.';
      }
      break;    
			
		case "autocompute":         
      msg.textContent= e.checked ? "The final result of the entered arithmetic expression will be calculated automatically at each change. This significantly simplifies the task." : "Off";                       
      break; 		
						
		case "is_integer":
			msg.textContent= e.checked ? 'Yes. The final result of the arithmetic expression must be an integer.' : 'No. The final result of an arithmetic expression can be a decimal number or an integer.';                       
      break; 	
		
    case "is_req":		
      msg.textContent= e.checked ? 'Yes' : 'No';      
      break; 
		
    case "warning":         
        msg.textContent= e.checked ? 'The user will need to confirm before the next task is displayed.' : 'The next task is displayed automatically, after clicking the button.';                       
      break; 				
		
		case "demo": 		
      msg.textContent= e.checked ? 'Yes' : 'No';      
      document.getElementById("sol_cnt").disabled= e.checked ? false: true; 
			break; 
				
    case "timer_enabled":   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById("timer").disabled= false;
      }
      else  {  
        msg.textContent= 'No';
        document.getElementById("timer").disabled= true; 
      }       
      break;
    
		case "timer":
      if (e.validity.valid)  {
        msg.textContent= stringTime(e.value); 
        msg.removeAttribute("class");
      }
      else  {
        msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
        msg.setAttribute("class", "error");
      } 
      break; 
		
    case "prg_type":                                                                                                                                            
      if (e.value === '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.';
      }
      else if (e.value === '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed.';
      }
      else  {
        msg.textContent= ''; //"Nothing";
      }
      break;    
      
    case "store":                                                                                                                                            
      msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
      break;  
      
  }   // switch (my_id)
       
}


// refresh main validator message for user defined OPTIONS (feedback to user)
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById("is_valid").textContent=  (dirty===true) ? "All your settings have been successfully updated" : "";  
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#000000"); 
  }
  else  {
    document.getElementById("is_valid").textContent="Some of your options are incorrectly set. They were not updated. These settings are marked in red."; 
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#ff0000;font-weight:bold"); 
  } 	
}


// check for validity.valid object only in all input type="number" 
function optionsValid()  
{     
  var res= document.getElementById("item_cnt").validity.valid && document.getElementById("timer").validity.valid && document.getElementById("sol_cnt").validity.valid;       
  return res;                   
}


// start task 
function startTask(e)   
{       
  
  setPar();      
  setPage();  
  
  document.getElementById("result_wrapper").style.display="none";
  document.getElementById("end").style.display= "none"; 
  document.getElementById("start").style.display= "none"; 
	timer_enabled= (document.getElementById("timer_enabled").checked) ? true: false;  // timer is enabled or disabled
	document.getElementById("time_left").textContent= "";
	if (prg_type === "2") {
	  document.getElementById("footer").setAttribute("style", "display:none");  
	}	
	else  
	{	
    document.getElementById("footer").setAttribute("style", "display:block");
		//document.getElementById("progress").style.display= "block";       		
		if (prg_type === "0") {
		  document.getElementById("progress").className="progress-bar progress-bar-striped progress-bar-info";     // items/total		
		}
		else if (prg_type === "1")  { 
		  document.getElementById("progress").className= "progress-bar progress-bar-striped progress-bar-success";  // remaining time   
		} 	
		document.getElementById("progress").setAttribute("style", "width:0");  // reset progress bar width
	}
	if (demo)  {
		document.getElementById('show_solution').style.display= 'block';
		document.getElementById('demo_check').checked= false;   // always off/unchecked "Show the solution(s)" checkbox
	}
	else  {
		document.getElementById('show_solution').style.display= 'none';	
	}	
	document.getElementById("show_item").style.display= "block";  	
	
  counter= 0;       // reset timer counter
  is_task= true;    // finally "activate" timer  	
  item=0;
	
  showItem();     // prepare (compute) and show first item     
}


// set application variable parameters
function setVarPar()  
{      
  if (document.getElementById("item_cnt").validity.valid) {
		itemCnt= parseInt(document.getElementById("item_cnt").value, 10);  // Total number of items in tasks 
  }

  exact= (document.getElementById("exact").value === "1") ? true : false;	

  autocompute= document.getElementById("autocompute").checked ? true: false;  
	
  is_int= document.getElementById("is_integer").checked ? true: false;  
	
  is_req= document.getElementById("is_req").checked ? true: false;  
  
	warning= document.getElementById("warning").checked ? true: false; 	
	
	demo= document.getElementById("demo").checked ? true: false; 
	 
  if (document.getElementById("sol_cnt").validity.valid) {
		solCnt= parseInt(document.getElementById("sol_cnt").value, 10);  
  }
   
	timer_enabled= document.getElementById("timer_enabled").checked ? true: false;  
		
	if (document.getElementById("timer").validity.valid) {
	  timer= Math.round(parseInt(document.getElementById("timer").value, 10));  
	}
  prg_type= document.getElementById("prg_type").value;    
   
  store= ( document.getElementById("store_group").disabled === false && document.getElementById("store").value === "1") ? true : false;  	
}


// set internal application parameters
function setPar()  
{  		
	dv= []; //  dv = dependent variable for each item=> two-dimensional array (matrix). 
  for (var i=0; i<itemCnt; i++)	
	{
		dv[i]= [];  // init default dependent variable values for each item..
		dv[i][0]= false; // [0] => false = arithmetic expression is not ok
		dv[i][1]= '';	// [1] => user entered arithmetic expression (string)
		dv[i][2]= 'Invalid';  // [2] => result of user arithmetic expression
		dv[i][3]= ''; // [3] => The best (optimal) result
		dv[i][4]= 0;    // [4] => Accuracy - percentage of similarity (absolute value) with the best solution 
		dv[i][5]= 100;  // maximum error = 100%. [5] => Error percentage (deviation - absolute value from the optimal result)	
	}
}


// set progress bar visible                                                                                    
function setPage()  { 
	document.getElementById("footer").style.display="block";     	
}


// get random integer from min to max range
function getRndInt(min, max) {
  return Math.floor(Math.random() * (max - min + 1)) + min;
}


// application timer handler function
function wm13Timer()  
{ 
  if (!is_task) {
		return;  // nop, idle, do nothing        
  }	       
	if (timer_enabled) 
	{
		counter++;
		if (counter >= timer)  {     // task time is limited
			theEnd();
			return;
		}	
		if (prg_type === "1")   {   // "1" => Time. display time on Progress Bar  
			var rest=timer-counter;   // this is remaining time 
			var percent= Math.round(100*rest/timer);    
			document.getElementById("progress").setAttribute("style", "width:"+percent+"%");    // !important
			document.getElementById("time_left").textContent= stringTime(rest);    
		}
	}			
} 


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  var hh = Math.floor(total/(60*60));         
  var divisor_for_minutes = total % (60 * 60);
  var mm = Math.floor(divisor_for_minutes / 60);       
  var divisor_for_seconds = divisor_for_minutes % 60;
  var ss = Math.ceil(divisor_for_seconds);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// prepare and show new item (task)
function showItem() 
{      
	document.getElementById('answer').value= '';
		
	compute();  // set (prepare) new task	
   
	// set (update) footer next item button (right arrow)
	var btn= document.getElementById("next");  
	if (item === itemCnt-1)  {
		btn.style.backgroundImage= "url('img/end.png')";
	  btn.title= "End of task";
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')";  
		btn.title= "Next task";
	}
	btn.style.display= "inline";  
  			
	// display (update) item count on Progress Bar (default)  
  if (prg_type === "0" )  {                                                                                                                                                         
    var percent = 100*(itemCnt-(item+1))/itemCnt;  // Math.round();      
    document.getElementById("progress").setAttribute("style", "width:"+percent+"%");    // !important
    document.getElementById("time_left").textContent= "Item= " + (item + 1) + " Remaining= " + (itemCnt - item-1); 		
  } 	
}


// input argument: dig = array of 5 different digits in range 1..9
// return array: arr[0]= numeric result of an arithmetic expression with [+ - * -] numerical operators
// return array: arr[1] = an arithmetic expression (string) 
function rndSol(dig)
{	
  var oper=[], res=[], s;			
	do
	{		
	  oper=[];         // we must create new operator array (do-while condition)
		oper[0]= 0;      // one more operator => oper[0] is not used	
		res[0]= dig[0];  // init numeric result with first digit
		s= dig[0].toString();    // an arithmetic expression (string) 
		for (var i=1; i<5; i++) 
		{ // iterate 4 digits (1..4)
			oper[i]= getRndInt(1,4);
			switch (oper[i])
			{
				case 1:
					res[0]= res[0] + dig[i]; 
					s+= "+" + dig[i]; 
					break;
				
				case 2:
					res[0]= res[0] - dig[i]; 
					s+= "-" + dig[i]; 
					break;
				
				case 3:			
  				res[0]= res[0] * dig[i]; 									
          if (i > 1 && (oper[i-1] == '1' || oper[i-1] == '2'))  {						
						s= '('  + s + ')*' + dig[i];             				
					}        
					else  {
					  s+= '*' + dig[i];
					}
					break;
				
				case 4:
					res[0]= res[0] / dig[i]; 
					
					if (i > 1 && (oper[i-1] == '1' || oper[i-1] == '2'))  {						
						s= '('  + s + ')/' + dig[i];             				
					}        
					else  {
					  s+= '/' + dig[i];
					}								
					break;
			}
		}
		
	} while ( !isInt(res[0]) || res[0] < 100 || res[0] > 999 || res[0] == target_num );  // range allowed: 100 .. 999. Target number ALWAYS must be integer	
	res[1]= s;
	return res;
}


/* Generates a task (item), 
   calculates all possible results and combinations between numbers, 
   finds the optimal arithmetic expressions and the corresponding numerical results */
function compute()
{	
	var rnd, sel={}, arr=[], target, e, ind, tmp, res2={}, res3={}, res4={}, res5={};  // res2..res5 => objects for intermediate results (tmp) 	
	window.result= {};
	for (var i=0; i<5; i++) 
	{
		do  {
			rnd= getRndInt(1,9);    
		} while (typeof sel[rnd] != "undefined");
		sel[rnd]= rnd;  // exclude this digit from next iterations
		arr.push(rnd);
  } 
	e= document.getElementById('item_digits');
	e.textContent='';
	for (var i in arr)  {
		e.textContent+= ' ' + arr[i] + ' ';  
	} 
	window.digits= arr;	
	target= rndSol(arr);	  	
	tmp= Math.round(0.1*target[0]); // add +-10% offset if there is no exact solution (exact == false)
	do  {
	  rnd= exact ? target[0] : getRndInt(target[0]-tmp, target[0]+tmp);  // set target_num for exact solution or approximate solution. target_num = target number
	  rnd= Math.round(rnd); // target number is always integer
	} while (rnd == target_num || rnd < 100 || rnd > 999 ); 
	target_num= rnd; // target_num => this is target number visible to user
	document.getElementById('user_target').textContent= target_num; 
  document.getElementById("user_res").textContent= ''; // clear result for previous item	
	//if (exact) {  optimal= target[1]; }	
	 
	for (var i=0; i<4; i++)  
	{
		for (var j=i+1; j<5; j++)  {			
			res2[ arr[i].toString() + arr[j].toString()]= get6(arr[i],arr[j]);   	
		}			
	} 
	
	for (var key1 in res2)  
	{  
		for (var i=0; i<5; i++)  
		{	
			if (key1.toString().indexOf(arr[i]) == -1) 
			{	
				ind= key1.toString() + arr[i].toString();
				res3[ind]=[]; // insert new empty array
				for (var j in res2[key1] )  {  		
					res3[ind].push(get6( res2[key1][j], arr[i] ));   							
				} 
			}
		}		
	}		

	for (var key1 in res3)  
	{  
		for (var i=0; i<5; i++)  
		{	
			if (key1.toString().indexOf(arr[i]) == -1) 
			{				
				ind= key1.toString() + arr[i].toString();
				res4[ind]=[]; 
				for (var j in res3[key1] ) {  
					for (var k in res3[key1][j] ) {
						res4[ind].push(get6(res3[key1][j][k], arr[i])); 
					}
				}
			}
		}		
	}		
    
	for (var key1 in res4)  
	{  
		for (var i=0; i<5; i++)  
		{	
			if (key1.toString().indexOf(arr[i]) == -1) 
			{				
				ind= key1.toString() + arr[i].toString();
				res5[ind]=[]; 
				for (var j in res4[key1] )  
				{
          for (var k in res4[key1][j] ) {								
						res5[ind].push(get6(res4[key1][j][k],arr[i]));  						
					}
				}
			}
		}		
	}		

	if ( is_int )  {
		difIntResult(res5);
	}
  else  {
		difResult(res5);
	}		
	var sol= getSolutions(target_num,solCnt); // Get best (or optimal) numeric results and arithmetic expression for this item
	dv[item][3]= sol[0][0]; // best solution for this item
	document.getElementById('solution').innerHTML='';
	if (demo)
  { 		
		document.getElementById('solution').innerHTML+= solCnt + ' best sorted solutions:';
		for (var i in sol)  {				
			if (i == 0) {
				document.getElementById('solution').innerHTML+= '<br>' + setExpression(sol[i][0], sol[i][1], true);  // save this 1. arithmetic expression (result) as optimal
			}
			else  {
				document.getElementById('solution').innerHTML+= '<br>' + setExpression(sol[i][0], sol[i][1], false);
			}
		}			
		if ( typeof window.result[target_num] != 'undefined' && typeof window.result[target_num][0] != 'undefined' )  {                                
			document.getElementById('solution').innerHTML+= '<br><b>' + 'There is an exact solution: ' + target_num + '</b>';  // + target[1] + ' = '  + ' optimal = ' + optimal + ' => ' + target_num + '</b>';     //  Exact solution:
		}
		else {
			document.getElementById('solution').innerHTML+= '<br><b>There is no exact solution.</b>';
		}
		document.getElementById('solution').innerHTML+= '<br>The displayed numbers can be calculated in several ways.'; // These are examples. 
	
	  showSolution(document.getElementById('demo_check'));
	}						 			
	document.getElementById('feedback').innerHTML= '';  // clear feedback (validator) message	
	document.getElementById('answer').focus();

}	

 
/* Calculates 6 numerical results from 2 numbers
	 n1, n2 = first and second number (digit).
	 Returns a 1-dim array with 6 possible results by performing basic arithmetic operations
*/
function get6(n1,n2)
{
	var res=[];
	//n1= parseInt(n1, 10); n2= parseInt(n2, 10);
	res.push(n1+n2);
	res.push(n1-n2);
	res.push(n1*n2);
	res.push(n1/n2);	
	res.push(n2-n1);
	if (n1 != 0) res.push(n2/n1); 
	return res;
}	


// filter all results into different results
// only integer results are accepted
function difIntResult(obj)
{
  for (var i in obj)  
	{
		for (var j in obj[i])  
		{	 
			for (var k in obj[i][j])  {
			  if (isInt(obj[i][j][k])) {
					window.result[ obj[i][j][k] ]= [ obj[i][j][k], i ];  //	index 0 = numeric result, index 1 = 5 digits combination	
				}
			}  
		}
	}
}


// filter all results into different results. 
// decimal results are accepted (and integer => all fynal different results)
function difResult(obj)
{
  for (var i in obj)  
	{
		for (var j in obj[i])  
		{	 
			for (var k in obj[i][j])  {			  
				window.result[ obj[i][j][k] ]= [ obj[i][j][k], i ];  //	index 0 = numeric result, index 1 = 5 digits combination					
			}  
		}
	}
}


/*finds the required value in a global object that contains a list of all stored results for current item
  target => target number to find (second index = 0)
	return 1 if found target number else return 0  */
function findResult(target)
{
	for (var key in window.result)  {		
		if  (window.result[key][0] == target) return 1;	
	}
	return 0;
}

	            
/* Creates an arithmetic expression for a given target number and the order in which 5 digits appear.  Adds parentheses if needed
   target = target number
	 digits = the order in which the 5 digits appear
	 save == true => save this arithmetic expression (result) as optimal 
	 These arguments come from the getSolutions () function */
function setExpression(target, digits, save)  
{
	var arr=[], res2=[], res3=[], res4=[], res5=[], op2=[], op3=[], op4=[], op5=[], tmp, tmp_val;   
	
	for (var i=0; i<5; i++)  {
	  arr[i]= parseInt(digits[i], 10);	
	}	
  
	res2.push(get6(arr[0], arr[1]));
	op2.push([0,1,2,3,4,5]);	
	
	for (var i in res2)  {
		for (var j in res2[i])  {  		
			res3.push( get6(res2[i][j], arr[2]) ); 	    					
			op3.push( addExp(op2[i][j]) );
		}	
	}	
   	
	for (var i in res3)  {
		for (var j in res3[i])  {  		
			res4.push( get6(res3[i][j], arr[3]) ); 
      op4.push( addExp(op3[i][j]) );			
		} 
	}
	
	for (var i in res4)  {
		for (var j in res4[i])  {  		
			res5.push( get6(res4[i][j], arr[4]) ); 
      op5.push( addExp( op4[i][j]) );  		
		} 
	}
	
	for (var i in res5)  
	{ 
		for (var j in res5[i])  
		{	
			if ( res5[i][j].toFixed(3) == target.toFixed(3) ) 
			{  
  			tmp= formatExp(digits,op5[i][j]);
				if (save) {
					optimal=tmp;
				} 
				tmp_val= parseExp(tmp);
				tmp+= isInt(tmp_val) ? '=' + tmp_val.toFixed(0) : '=' + tmp_val.toFixed(3);			// tmp+= "=" + parseExp(tmp);								
				return tmp;
			}
		}
	} 	
	return ""; // if error => show empty line
}


/*just adds internal flag: 0..5 to a string 
  exp = coded string (internal application labels/flags for operators. exp = operators array)
  0 = '+', '1' = '-', 2 = '*',.. see formatExp() function for decoding */ 
function addExp(exp)
{
	var res=[];
	for (var k=0; k<6; k++)  {
	  res.push( exp.toString() + k.toString() );
	}  
	return res;
}	


/* creates (formats) an arithmetic expression for a given string of digits and operators. Adds parentheses if needed
  d = 5 digits. The order in which 5 digits appear when creating a numeric result
  op = operators. range 0..5 => see functions get6() and addExp()   */
function formatExp(d,op)
{
	/*  function get6()
	res.push(n1+n2); 0
	res.push(n1-n2); 1
	res.push(n1*n2); 2
	res.push(n1/n2); 3
	res.push(n2-n1); 4
	if (n1 != 0) res.push(n2/n1);  5	 */
  
	var s= d[0];
	for (var i=0; i<4;  i++)  
	{ // operands are in range 0..3	
		switch (op[i])
		{
			case '0': 
				s+= '+' + d[i+1];
				break;  			
			case '1': 
				s+= '-' + d[i+1];	
				break;
			case '2': 
				if ( i > 0 && (op[i] == '2' || op[i] == '3' || op[i] == '5') && (op[i-1] == '0' || op[i-1] == '1' || op[i-1] == '4' ) )  {
					s= '(' + s + ')' + '*' + d[i+1];  // add new bracket
				}
				else  {
					s+= '*' + d[i+1];	
				}
				break;
			case '3': 
				if ( i > 0 && (op[i] == '2' || op[i] == '3' || op[i] == '5') && (op[i-1] == '0' || op[i-1] == '1' || op[i-1] == '4' ) )  {
					s= '(' + s + ')' + '/' + d[i+1];   // add new bracket
				}
				else  {
					s+= '/' + d[i+1];	
				}
				break;
			case '4': 
				s= i > 0 ? d[i+1] + '-(' + s + ')' : d[i+1] + '-' + s;  // d[i+1] - (d[i]) => always add new bracket '()' if i>0
				break;
			case '5': 
				s= i > 0 ? d[i+1] + '/(' + s + ')' :  d[i+1] + '/' + s;	   
				break;	
      default:
				s+='?'; // error
		}		
	}
	return s;
}


/* return arguments for setExpression(target, digits) function    
   target => best solution = target
   n => number of return solutions. n must be integer >= 1 
	 return 2-dim array of numbers closest to the target number. 
	 for each number: index 0 => numbers value, index 1 => The order in which 5 digits appear when creating a numeric result  */
function getSolutions(target, n)
{
	//if (n < 1) return;
	var key, dif, pos=[], neg=[], res=[];  // max (int) for z-index:2147483647;
	for (var key in result)  
	{  
    dif= result[key][0] - target; 
		if (dif >= 0)  {
		  pos.push([Math.abs(dif), result[key][1]]);  // [ 0 = abs difference,  1= The order in which 5 digits appear when creating a numeric result ]
		}	
		else  {
			neg.push([Math.abs(dif), result[key][1]]);  // [ 0 = abs difference,  1= The order in which 5 digits appear when creating a numeric result ]
		}	
	}  	
	//https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/sort
	// https://riptutorial.com/javascript/example/3443/sorting-multidimensional-array
	// https://stackoverflow.com/questions/16096872/how-to-sort-2-dimensional-array-by-column-value
	// use numeric sort
	pos.sort(function(a, b) {
		return a[0] - b[0];
	});
	neg.sort(function(a, b) {
		return a[0] - b[0];
	});
	
	var pos_len= pos.length, neg_len= neg.length, pos_j=0, neg_j=0; 
	for (var i=0; i<n; i++)  
	{
		if (pos_j < pos_len && neg_j < neg_len)  
		{
      if ( pos[pos_j][0] <= neg[neg_j][0]) {
				res.push( [ target + pos[pos_j][0], pos[pos_j][1] ]);
				pos_j++;
			}
			else  {
			  res.push( [ target - neg[neg_j][0], neg[neg_j][1] ] );		
				neg_j++;
			}
		}
		else if (pos_j < pos_len && neg_j >= neg_len)  {
      res.push( [ target + pos[pos_j][0], pos[pos_j][1] ] );			
			pos_j++;
		} 
		else if (neg_j < neg_len && pos_j >= pos_len)  {
      res.push( [ target - neg[neg_j][0], neg[neg_j][1] ]);		
			neg_j++;
		} 	
	}  	
	return res;  // for each number: index 0 => numbers value, index 1 => the order in which 5 digits appear when creating a numeric result  
}


/* 	checks user arithmetic expression
		return true if arithmetic expression is valid
		else return false */ 
function check()  
{		
	var res= validate();
	//if (res[0] == 'ok') => the entered arithmetic expression is valid. In that case: res[1]= calculated value of a valid arithmetic expression
	//if (res[0] != 'ok') => the entered arithmetic expression is invalid. It then contains a list of errors within the entered arithmetic expression					
	if (res[0] == 'ok') 
	{    
		if (autocompute) document.getElementById('user_res').textContent= ' = ' + res[1];
		if (findResult(res[1]) == 1 )  {
			document.getElementById('feedback').textContent= ' Solution was accepted.';			
		}
		else  {	  
			document.getElementById('feedback').textContent= ' New solution was accepted.'; // A new solution has been found. Not found. 
		}
		return true;
	}
	else  
	{ 
		var e= document.getElementById('feedback');  // https://stackoverflow.com/questions/18990441/innerhtml-br-breaking
		e.innerHTML='Errors: <br>'; // The following errors were found:       The entered arithmetic expression is invalid.
		for (var i in res) {
			e.innerHTML+= res[i] + '<br>';		
		}
		document.getElementById('answer').focus();
		return false;
	}
}


/* validate (and parse) user arithmetic expression (string)
   If the entered arithmetic expression is correct, it calculates the numerical value of the arithmetic expression.
	 Otherwise, it returns the errors found within the arithmetic expression (if any)
   return err[] array with errors or 'ok' flag (see the source code below for a more detailed explanation)
	 if (err[0] == 'ok') => the entered arithmetic expression is valid. In that case: err[1]= calculated value of a valid arithmetic expression
	 if (err[0] != 'ok') => the entered arithmetic expression is invalid. It then contains a list of errors within the entered arithmetic expression */ 
function validate()
{ 
  var err=[], op= 0, tmp, b=false;  // err => function result
  var e= document.getElementById('answer');
	var s= trim(e.value); 	
	//s= s.replace('^', '');
	//s= s.replace('**', '*'); // s=> js syntax ** = ^
	s= s.replace(' ', '');
	s= myTrim(s);
	document.getElementById('answer').value=s;	
	for (var i= 0; i< s.length; i++)  
	{ // each substring can be only: digit "0..9" or oper "+-*/" or block "(") 
	  if ( isDigit(s[i]) ) {
			if ( !validDigit(s[i]) ) err.push('Invalid digit: ' + s[i]);			
		}
		else if ( isOper(s[i]) )  {
			op++;		// must check number of operators. else for target 123 this be valid: 123 ; 123*1 ; 123+0; 123*1*1+0
      if ( i > 0 && isOper(s[i-1]) && b === false) {
				err.push('Invalid operator: you can only use: +, -, *, /');		// only one warning
	      b=true;
				//op--;
			}
		}	
		else if ( !isBlock(s[i]) )  {
			err.push('Invalid character: ' + s[i]);
		}					
	}
	var tmp= allDigits(s);
	if (tmp.length > 0)  
	{
	  for (var i in tmp) { 
		  err.push('You must use digit:' + tmp[i]); 
		} 
	}
  tmp= onlyOnce(s);
	if (tmp.length > 0)  
	{
	  for (var i in tmp) { 
		  err.push('This digit is used more than once:' + tmp[i]); 
		} 
	}
	if (op != 4)  {
		err.push('Invalid number of operators (+-*/): ' + op);		
	}
		
	if (err.length === 0)  
	{			
		try  {
		  tmp= parseExp(s); // numeric result of arithmetic expression
		}
		catch(e) {
			err[0]='NaN';
		}		
	  if (isNaN(tmp)) {
			err[0]='NaN';
		}
		else if (is_int && (!isInt(tmp)) ) {
			err[0]='The result of an arithmetic expression must be an integer';  
		}			
		else  {		
			err[0]= 'ok';  // arithmetic expression for the current item (string) is valid
			err[1]= tmp; // numeric result of arithmetic expression (NOT rounded to 3. decimals)			
		}
	}		
	return err;	
}


/* see function init().  deprecated but still works:
   https://developer.mozilla.org/en-US/docs/Web/API/GlobalEventHandlers/onkeypress
   this regex [1-9()*\/+-] was tested here:    https://regex101.com/    */
function answerKeyPress(event)   { 
  return event.charCode === 0 || /[1-9()*\/+-]/.test(String.fromCharCode(event.charCode));	
}


/// when the entered arithmetic expression changes. see wm13.html tag
function answerInput(e)  
{	
	if (autocompute)  
	{
		var tmp= e.value;
		if (trim(tmp) == '') 
		{
			document.getElementById("user_res").textContent=''; 
			return; 
		}
		//document.getElementById("user_res").textContent= ' Invalid value'; 		
		try  {
			document.getElementById("user_res").textContent= isInt(tmp) ? ' = ' + parseExp(tmp).toFixed(0) : ' = ' + parseExp(tmp).toFixed(3); 	 
		}
		catch(e) {
			document.getElementById("user_res").textContent= ' Invalid value';
		}
	} 
}


// go to the next task if the ENTER key is pressed. see function init()
function answerKeyUp(e)  
{
	if (e.key === 'Enter' || e.keyCode === 13) {  
		nextItem();
	}
}


// parse arithmetic expression (string)
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/eval
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Function
function parseExp(obj){
	try  {
		return Function('"use strict";return (' + obj + ')')();		
	}
	catch(e) {
		return "Invalid";
	}	
}


// trim string
function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}


// Checks whether it is a digit or not 
function isDigit(s) 
{
	if (s.length != 1) return false;
	var digit= { 0:0, 1:1, 2:2, 3:3, 4:4, 5:5, 6:6, 7:7, 8:8, 9:9 };
	if (digit[s]) {
		return true;
	}
	else  {
		return false;
	}
}


// Checks if the digit is valid or not	
function validDigit(s)  {
  if(window.digits.length != 5 || s.length != 1) return false;
  return ( s == window.digits[0] || s == window.digits[1] || s == window.digits[2] || s == window.digits[3] || s == window.digits[4] );
}


// check if all window.arr digits are used in calculated string
function allDigits(s)  {
  var res=[];  //if (window.digits.length != 5) return "Error";
	for (var j in window.digits)  {
		if ( s.indexOf(window.digits[j]) === -1 ) {
			res.push( window.digits[j] ); 			
		}
	}
	return res;
}


// check if all window.arr digits are used only once in calculated string
function onlyOnce(s)  {
  var res=[], cnt;  //if (window.digits.length != 5) return "Error";
	for (var i in window.digits)  
	{
		cnt=0;
		for (var j in s) {			
		  if (s[j] == window.digits[i]) cnt++;
		}
		if (cnt > 1) res.push( window.digits[i] ); 		
	}
	return res;
}


// Checks if it is a numeric operator: +, -, *, /
function isOper(s)  {
	//if  (s.length != 1) return false;
	return ( s == '+' || s == '-' || s == '*' || s == '/' );
}	


// Checks parentheses within an arithmetic expression
function isBlock(s)  {
	return ( s == '(' || s == ')' );
}	


// trim string  + delete all spaces within string ' ' => ''
function myTrim(s) {
	var res= '';
  for (var i=0; i< s.length; i++)  {
		if (s[i] != ' ') res+= s[i];
	}
	return res;
}


// Checks whether it is an integer value or not
// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


// Displays a list of the window.solCnt best solutions generated by this application (sorted from best to worst)
// Sets the best arithmetic expression in the input field (answer)
function showSolution(e) 
{
 if (e.checked)  {
	 document.getElementById('solution').style.display="block";
	 document.getElementById('answer').value= optimal;  // Sets the best arithmetic expression in the input field (answer)
	 var tmp= parseExp(optimal);
	 document.getElementById("user_res").textContent= isInt(tmp) ? ' = ' + tmp.toFixed(0) : ' = ' + tmp.toFixed(3);
 }
 else  {
   document.getElementById('solution').style.display="none";
	 document.getElementById('answer').value='';	
	 document.getElementById("user_res").textContent= ''; 
 }	 
}  


// goto next item 
function nextItem()  
{ 
  var res= check();  // this return if arithmetic expression is valid (true) or not (false) 	
	if (!res && is_req)  {  // a non-integer value is accepted if is_req = false
		document.getElementById('feedback').innerHTML+= '<p class="error">You must enter a valid arithmetic expression before moving on to the next task</p>';
		document.getElementById('answer').focus();	
		return;   // the user must enter a valid arithmetic expression before moving on to the next task
	}	
	
	if (warning) {		
	  var s= (item < itemCnt-1) ? 'Going to the next task? \n Warning: After moving on to the next task, it is no longer possible to return to this task.' : 'End of task? \n  Warning: This action cannot be undone.';
		if ( !window.confirm(s) ) { 
			return;
		}  
	}
	/* save user response here
    dv = [] = dependent variable for each item=> two-dimensional array (matrix). 2. index explanation:  
    [0] => true => arithmetic expression is valid;  false => arithmetic expression is not ok
		[1] => user entered arithmetic expression (string)
		[2] => user result of arithmetic expression
		[3] => The best (optimal) result
		[4] => Accuracy - percentage of similarity (absolute value) with the best solution
		[5] => Error percentage (deviation - absolute value from the optimal result)	*/	
	dv[item][0]= (res === true) ? true : false;	// arithmetic expression is valid or not
	dv[item][1]= trim(document.getElementById('answer').value);	// user entered arithmetic expression (string)
	dv[item][2]= (dv[item][0] === true) ? parseExp(dv[item][1]) : 'Invalid';  // user result of arithmetic expression (number) or string 'Invalid' 	 
	if (dv[item][0] === true) 
	{
		// dv[item][3]=  => default value => best (optimal) result is set in compute() => solutions generated by the application
		// Check if the user offered a solution better compared to the solutions generated by this application
		if ( !isNaN(dv[item][2]) && ( Math.abs(parseFloat(dv[item][2]) - target_num) < Math.abs(parseFloat(dv[item][3]) - parseFloat(target_num)) ) )  {
			dv[item][3]= dv[item][2];  // The user's solution becomes the best (optimal) if it is better than those generated by this application 
 		}				
		//dv[item][4]= (100 * (1-Math.abs( parseFloat(dv[item][3]) - parseFloat(dv[item][2]) ) / parseFloat(dv[item][3]))).toFixed(3);  // Accurancy %
		dv[item][5]= (100 * Math.abs( parseFloat(dv[item][3]) - parseFloat(dv[item][2]) ) / parseFloat(dv[item][3])).toFixed(3); // Error %	
		// force accuracy and error in the range: 0%-100%
		dv[item][5]= Math.abs(dv[item][5]);  
		if (dv[item][5] > 100) { dv[item][5]= 100; }  // error is always positive in range 0% - 100%	
		dv[item][4]= (100-dv[item][5]).toFixed(3);   // accurancy is always positive in range 0% - 100%
	}
	else  {
		dv[item][4]= 0;    // minimum accurancy = 0% 
		dv[item][5]= 100;  // maximum error = 100% 
	}
	if (item < itemCnt-1)  {         
    item++; 
    showItem();     // show new item        
  }
  else  {               
    theEnd();  // end of task  
  }
}


// end of task (research)
function theEnd()  
{   
  is_task= false;  
  timer_enabled= false; 
	
	//document.getElementById("footer").style.display="none";
  document.getElementById("end").style.display="block";  
  document.getElementById("show_item").style.display= "none"; 
  
	// override is_valid column index = [0]. use is_valid column for item number
  for (var i=0; i< dv.length; i++)  {
	  if (dv[i][0] === true)  {
			dv[i][2]= isInt(dv[i][2]) ? dv[i][2].toFixed(0) : dv[i][2].toFixed(3);  // if valid round user result to max 3. decimals. [2] => user result of arithmetic expression
			dv[i][3]= isInt(dv[i][3]) ? dv[i][3].toFixed(0) : dv[i][3].toFixed(3);  // round best result to max 3. decimal
		}				
	}		
				
  document.getElementById("result_wrapper").style.display="block";  
	createTableArray(document.getElementById("table1"), dv,  ['Item', 'Arithmetic expression' , 'Result', 'Best result', 'Accuracy %', 'Error %']);  
	document.getElementById("table1").style.display="block";  
	
	// calculate decriptive statistics for accurancy and errors (average and standard deviation)..
  var tmp_c=[], tmp_e= [], inv= 0, mis= 0, m_c, sd_c, m_e, sd_e, tmp=[];
	for (var i=0; i< dv.length; i++)  {
		tmp_c[i]=parseFloat(dv[i][4]);  // Percentage of accuracy 
    tmp_e[i]=parseFloat(dv[i][5]);  // Percentage of error 
		if ( dv[i][0] != true ) { inv++; }	  // inv => invalid answers
    if ( trim(dv[i][1]) == '' ) { mis++; }  // mis => missing aritmetic expression (answers) => no answer			
	  dv[i][0]= i+1;  // at the end => only for print table: the validity column (index 0) now becomes the item number
	}			
	m_c= mean(tmp_c).toFixed(3); // mean accurancy
	sd_c= (itemCnt > 1) ? stDev(tmp_c, m_c).toFixed(3) : 'Cannot be calculated'; // standard deviation accurancy
  m_e= mean(tmp_e).toFixed(3);  // mean errors
	sd_e= (itemCnt > 1) ? stDev(tmp_e, m_e).toFixed(3) : 'Cannot be calculated'; // standard deviation errors
	//tmp[0]= [];
	tmp[0]= ['Accuracy', m_c, sd_c, itemCnt];
	//tmp[1]= [];
	tmp[1]= ['Error', m_e, sd_e, itemCnt];			
	createTableArray(document.getElementById("table2"), tmp, ['Variable', 'Average', 'Sample Standard Deviation', 'N']); 
  document.getElementById('chart_title').textContent= 'Accuracy percentage (green) = ' + m_c + ' Error percentage (red) = ' + m_e;
	document.getElementById("score").style.display= store ? "block" : "none";
	try 
  {
    if (store) {
      drawStorage(m_c,sd_c,inv,mis);  
    }
  }        
  catch (e)  { 
    document.getElementById("table_grp").textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
  // draw - repaint "chart" 
	document.getElementById("chart_cor").setAttribute("style", "width:"+m_c+"%");    // !important
	document.getElementById("chart_err").setAttribute("style", "width:"+m_e+"%");
}


// return average (mean) of array (arr) 
function mean(arr)  {
	var sum=0;
	for (var i in arr)  {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation
// arr = array argument; mean = average (mean)
function stDev(arr, mean) {
	var len= arr.length, sum= 0, res;
	for (var i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow( sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function
function setDate()  
{
	var today = new Date();
	var dd = today.getDate();
	var mm = today.getMonth()+1; 
	var yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/*  use window.localStorage in client browser
  m = Average (Mean) for Accuracy percent (%)
	sd = Sample Standard Deviation for Accuracy percent (%)
	inv = total number of invalid Arithmetic expressions (results)
	mis = total number of missing answers (The arithmetic expression is not entered at all)
		 */			
function drawStorage(m, sd, inv, mis)   
{  	
	document.getElementById("table_grp").style.display="block";
  var oldItems = JSON.parse(localStorage.getItem("wm13-v1.0")) || []; 
  var id= oldItems.length+1;   
  // set new "record" 
  var newItem = {
    'id': id,
    'date': setDate(),    //new Date(),          
		'M-accuracy': m, 
    'SD-accuracy': sd,       
    'missing': mis,  
		'invalid': inv,
    'show-correct' : demo ? "1": "0", 	 
    'calculator' : autocompute ? "1" : "0",
		'is_exact': exact ? "1" : "0",
		'is-integer': is_int ? "1" : "0",
		'required': is_req ? "1" : "0",
		'warning': warning ? "1" : "0",
		'items': itemCnt,  
    'timer-s' : document.getElementById("timer_enabled").checked ? timer : "", 
		'progress':  document.getElementById("prg_type").value                       		
  };

  oldItems.push(newItem);

  localStorage.setItem("wm13-v1.0", JSON.stringify(oldItems));

  createTableJSON(document.getElementById("table_grp"), oldItems); 
   
  document.getElementById("score").textContent= "The results are stored in your browser. (ID=" + id + ")";
 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  var col = [];
  for (var i = 0; i < my_json.length; i++) {
    for (var key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  var table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  var tr = table.insertRow(-1);                   // TABLE ROW.
  for (var i = 0; i < col.length; i++) {
      var th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (var i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (var j = 0; j < col.length; j++) {
      var tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert two-dimensional array (matrix) data dynamically to HTML table - dynamically create html table with rows
   elem == DOM html element (node)
   data => 1dim or 2-dim array (matrix)
   col => String column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	var table = document.createElement("table");  // html table
  var tr = table.insertRow(-1), th, cell;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (var i = 0 in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th = document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{
		// add 2-dim array data to the table as rows
		for (var i = 0; i < data.length; i++) 
		{
			tr= table.insertRow(-1);
			for (var j in col) {
				if ( col[j] == '' ) { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell = tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// // add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		for (var i = 0; i < data.length; i++)	{
			cell = tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}
	// finally add the newly created table with json data to a container
  elem.innerHTML= "";
  elem.appendChild(table); 
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr){
    if(arr[0] === undefined){
        return false;
    }else{
        return (arr[0].constructor === Array);
    }
}


// clear wm13 web browser storage data
function clrStorage()  
{  
  var cnt= localStorage.getItem("wm13-v1.0") ? JSON.parse(localStorage.getItem("wm13-v1.0")).length : 0;                                                        
  if (cnt === 0)  {
    alert("You have no wm13 data to delete");
    return;
  }  
  if ( !window.confirm("You have a total of " + cnt  + " wm13 results stored in your web browser. Continue?") ) { 
   return;
 }                                             
 if (window.confirm("Delete total saved wm13 results:  "  + cnt + ". This action cannot be undone. Final confirmation?") ) 
 { 
   window.localStorage.removeItem("wm13-v1.0"); 
   document.getElementById("store_cnt").textContent= "All data has been deleted. No data";
 }  
}
